# SPDX-FileCopyrightText: © 2014 Josef Hahn
# SPDX-License-Identifier: AGPL-3.0-only

import os
import sys

import parsley.aspect.applypathacceptor
import parsley.aspect.collectinformation
import parsley.aspect.conflicts
import parsley.aspect.defaultiteration
import parsley.aspect.electmaster
import parsley.aspect.readmefile
import parsley.aspect.remove
import parsley.aspect.update
import parsley.aspect.volumepathbreadcrumb
import parsley.aspect.volumesyncreport
import parsley.syncengine.common
import parsley.tools.pathacceptors


class DefaultBaseBare(parsley.aspect.defaultiteration.DefaultIteration,
                      parsley.aspect.collectinformation.CollectInformation,
                      parsley.aspect.applypathacceptor.ApplyPathAcceptor,
                      parsley.aspect.readmefile.ReadmeFile,
                      parsley.aspect.volumepathbreadcrumb.VolumePathBreadcrumb,
                      parsley.aspect.volumesyncreport.VolumeSyncReport):
    """
    Very basic part of default behavior for a plain synchronization.
    Part of parsley.aspect.defaults.DefaultBase.
    """

    def __init__(self):
        parsley.aspect.defaultiteration.DefaultIteration.__init__(self)
        parsley.aspect.collectinformation.CollectInformation.__init__(self)
        parsley.aspect.applypathacceptor.ApplyPathAcceptor.__init__(self,
                                                                    parsley.tools.pathacceptors.builtinpathacceptor)
        parsley.aspect.readmefile.ReadmeFile.__init__(self)
        parsley.aspect.volumepathbreadcrumb.VolumePathBreadcrumb.__init__(self)
        parsley.aspect.volumesyncreport.VolumeSyncReport.__init__(self)


class DefaultBase(DefaultBaseBare,
                  parsley.aspect.electmaster.ElectMasterFileByMtime,
                  parsley.aspect.electmaster.ElectMasterLinkByTargetHistory,
                  ):
    """
    Mid-basic part of default behavior for a plain synchronization.
    Part of parsley.aspect.defaults.DefaultSync.
    """

    def __init__(self):
        DefaultBaseBare.__init__(self)
        parsley.aspect.electmaster.ElectMasterFileByMtime.__init__(self)
        parsley.aspect.electmaster.ElectMasterLinkByTargetHistory.__init__(self)


class DefaultSync(DefaultBase, parsley.aspect.update.DefaultUpdateItems, parsley.aspect.conflicts.DetectTypeConflicts,
                  parsley.aspect.conflicts.TrackConflicts, parsley.aspect.conflicts.ResolveConflictsByHint,
                  parsley.aspect.remove.DetectRemoval, parsley.aspect.remove.DefaultRemoveDirs):
    """
    Default behavior for a plain synchronization.
    """

    def __init__(self):
        DefaultBase.__init__(self)
        parsley.aspect.update.DefaultUpdateItems.__init__(self)
        parsley.aspect.conflicts.DetectTypeConflicts.__init__(self)
        parsley.aspect.conflicts.TrackConflicts.__init__(self)
        parsley.aspect.conflicts.ResolveConflictsByHint.__init__(self)
        parsley.aspect.remove.DetectRemoval.__init__(self)
        parsley.aspect.remove.DefaultRemoveDirs.__init__(self)


class PullAndPurgeSyncSource(DefaultBase):
    """
    Default behavior for a source filesystem in a pull-and-purge configuration.
    """

    @parsley.aspect.hook("defaultupdate", "", "",
                         event=parsley.syncengine.common.SyncEvent.UpdateItem_Update_ExistsInMaster)
    @parsley.aspect.execute_only_for_master_fs()
    @parsley.aspect.execute_only_for_master_fs_filetype(parsley.syncengine.common.EntryType.File)
    @parsley.aspect.execute_only_if_not_update_set_skipped()
    def ppsyncsource_removefromsource(self, ctx, filesystem):
        """
        Removes the file in the source filesystem.
        """
        ctx.delete_file(filesystem, ctx.path)


class PullAndPurgeSyncSink(DefaultBaseBare, parsley.aspect.update.DefaultUpdateItems,
                           parsley.aspect.remove.CleanupTrashBin):
    """
    Default behavior for a sink filesystem in a pull-and-purge configuration.
    """

    def __init__(self):
        DefaultBaseBare.__init__(self)
        parsley.aspect.update.DefaultUpdateItems.__init__(self)
        parsley.aspect.remove.CleanupTrashBin.__init__(self)

    @parsley.aspect.hook("", "", "", event=parsley.syncengine.common.SyncEvent.UpdateItem_ElectMaster)
    @parsley.aspect.execute_only_if_not_already_maximally_elected()
    def ppsyncsink_electifnowhereelse(self, ctx, filesystem):
        """
        Elects the sink filesystem if the item exists here ...
        - by a minimal key, if it is a file, so it gets elected if there are no others
        - by a maximum key, if it is a directory, so the sync can go deeper
        """
        if ctx.getinfo_current_exists(filesystem, ctx.path):
            key = sys.maxsize \
                if (ctx.getinfo_current_ftype(filesystem, ctx.path) == parsley.syncengine.common.EntryType.Directory) \
                else -1
            ctx.promote_master_filesystem(filesystem, key)

    @parsley.aspect.hook("", "", "defaultupdate", event=parsley.syncengine.common.SyncEvent.UpdateItem_CheckConflicts)
    @parsley.aspect.execute_only_if_not_update_set_skipped()
    @parsley.aspect.execute_only_for_non_master_fs()
    @parsley.aspect.execute_only_for_master_fs_filetype(parsley.syncengine.common.EntryType.File)
    def ppsyncsink_renameexistingtonew(self, ctx, filesystem):
        """
        Renames an already existing entry to some new name.
        """
        _p = ctx.path
        i = 1
        _pd = os.path.dirname(ctx.path)
        _pb = os.path.basename(ctx.path)
        _pi = _pb.rfind(".")
        rename = False
        while filesystem.exists(_p):
            rename = True
            if _pi > 0:  # sic!
                _p = "{pd}/{pb1}.{i}.{pb2}".format(pd=_pd, pb1=_pb[:_pi], i=i, pb2=_pb[_pi + 1:])
            else:
                _p = ctx.path + "." + str(i)
            i += 1
        if rename:
            filesystem.move(ctx.path, _p)
            ctx.setinfo_removefile(filesystem, ctx.path)
            # noinspection PyProtectedMember
            ctx.syncglobaldata.changed_flag = True
